/**==========================================================================

	PC-8031 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2024 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	last modified  Apr,27,2024.

;==========================================================================*/
#include "../PC-8001/PC-8001.h"
#ifdef __APPLE__
#include <string.h>
#endif
#include <crtdbg.h>
#ifdef __cplusplus
extern "C"
{
#endif
#define export
#define DIAGSIZE	(6250)
/**--------------------------------------------------------------------------
 **		Disk Drive
 **------------------------------------------------------------------------*/
static void _drv_eject(LPFDUDRIVE drv) {
	if (!drv->media_inserted)
		return;
	if (drv->media == NULL) {
		drv->media_inserted = FALSE;
		return;
	}
	drv->media->dispose(drv->media);
	drv->media_inserted = FALSE;
	drv->media = NULL;
}

static BOOL _drv_insert(LPFDUDRIVE drv, LPFLOPPYDISK media) {
	if (drv->media_inserted)
		return FALSE;
	if (drv->media != NULL) {
		drv->media_inserted = TRUE;
		return FALSE;
	}
	drv->media_inserted = TRUE;
	drv->media = media;
	drv->media->mounted = TRUE;
	drv->media->image->ref_count++;
	return TRUE;
}

static char _drv_get_state(LPFDUDRIVE drv) {
	return drv->missing_unit || !drv->media_inserted ? '-' : drv->state;
}

static void _drv_motor_control(LPFDUDRIVE drv, BYTE motor) {
	if (drv->motor != motor) {
		drv->motor = motor;
#if _DEBUG
		if (motor == 0)
			n80device.logger("FDU: [DRIVE #%d] Motor ON --> OFF\n", drv->unit_no);
		else
			n80device.logger("FDU: [DRIVE #%d] Motor OFF --> ON\n", drv->unit_no);
#endif
	}
}

static void _drv_set_mode(LPFDUDRIVE drv, BYTE td, BYTE rv) {
	drv->tpi	 = td;	// 0:48TPI,  1:96TPI
	drv->drvmode = rv;	// 0:2D/2DD, 1:2HD
}

static LPD88SECTOR_EX _drv_find_sector_in_track(LPFDUDRIVE drv, LPFDUCOMMAND cmd) {
	return drv->media->find_sector_in_track(drv->media, drv->cylinder, cmd->HD, cmd->MT, &drv->id, &drv->track, &drv->sector, &drv->header);
}
static BOOL _drv_forward_id(LPFDUDRIVE drv, LPD88SECTOR sector, BYTE MT, LPSECTORID id) {
	BOOL EOT = TRUE;
	int ns = -1;
	int sectors = drv->media->sectors(sector);
	for (int s = 0; s < sectors; ++s) {
		LPD88SECTOR next = (LPD88SECTOR)drv->media->get_sector(drv->media, mktrkidx(drv->track, s));
		if (next == NULL)
			continue;
		if (next->id.R > sector->id.R) {
			if (ns < 0 || id->R > next->id.R) {
				ns = s;
				*id = next->id;
			}
		}
	}
	if (ns >= 0)
		return FALSE;

	if (MT == 0) {
		id->C++;
		id->R = 1;
	} else {
		if (id->H == 0) {
			EOT = FALSE;
			id->H = 1;
			id->R = 1;
		} else {
			id->C++;
			id->H = 0;
			id->R = 1;
		}
	}
	return EOT;
}

static void _drv_reset(LPFDUDRIVE drv) {
	drv->status = 0;
	drv->state	= '#';
	drv->motor	= 0;	// motor off
	drv->busy	= DRVBSY_IDOL;
}

static void _drv_init_drive(LPFDUDRIVE drv, BYTE unit_no) {
	drv->reset			= _drv_reset;
	drv->eject			= _drv_eject;
	drv->insert			= _drv_insert;
	drv->get_state		= _drv_get_state;
	drv->motor_control	= _drv_motor_control;
	drv->set_mode		= _drv_set_mode;
	drv->forward_id		= _drv_forward_id;
	drv->find_sector	= _drv_find_sector_in_track;
	drv->unit_no		= unit_no;
	drv->media_inserted = FALSE;
	drv->media			= NULL;
	drv->reset(drv);
}
/**--------------------------------------------------------------------------
 **		Disk Drives
 **------------------------------------------------------------------------*/
static int _fdd_seek(LPFDUDRIVES fdd, int unit, int cylinder, int header) {
	LPFDUDRIVE drive = &fdd->drives[unit];
	int interval = 1;
	int pcn = drive->cylinder;
	if (drive->missing_unit /*|| !drive->media_inserted*/) {
		drive->status = (unit & FDC_ST0_US_MASK) | FDC_ST0_SE | FDC_ST0_NR;// | FDC_ST0_IC_ABNORM_TERM;
		interval = 16 - fdd->owner->fdc.step_rate_time;
	} else {
		drive->status = (unit & FDC_ST0_US_MASK) | FDC_ST0_SE;
		interval = 16 - fdd->owner->fdc.step_rate_time;
	}
	drive->cylinder = cylinder;
	drive->header	= header;
	drive->sector	= -1;
	drive->state	= 'S';
	int diff = trkdiff(cylinder, pcn);
	interval *= diff;	// 1..16ms
	return interval == 0 ? 120 : fdd->owner->msec_to_T(fdd->owner, interval);
}

static void _fdd_sense_interrupt_status(LPFDUDRIVES fdd) {
	LPFDU765A fdc = &fdd->owner->fdc;
	for (int i = 0; i < FDC_MAX_UNIT_CNT; ++i) {
		LPFDUDRIVE drive = &fdd->drives[i];
		if (drive->status != 0) {
			fdc->result.push0(&fdc->result, drive->status & ~FDC_ST0_HD);	// ST0
			fdc->result.push(&fdc->result, drive->cylinder);				// PCN
			drive->status = 0;
			drive->state = 'Q';
			return;
		}
	}
	fdc->result.push0(&fdc->result, FDC_ST0_IC_INVALID_CMD);	// ST0
	fdc->result.push(&fdc->result, 0);							// PCN
}

static void _fdd_sense_drive_status(LPFDUDRIVES fdd, int unit, int head) {
	LPFDU765A fdc = &fdd->owner->fdc;
	LPFDUDRIVE drive = &fdd->drives[unit];
	BYTE ST3 = (unit & FDC_ST3_US_MASK);
	if (drive->missing_unit)
		ST3 |= FDC_ST3_FT;
	else if (drive->media_inserted) {
		ST3 |= (FDC_ST3_RY | FDC_ST3_TS | (drive->cylinder == 0 ? FDC_ST3_T0 : 0) | (drive->media->header->readonly == 0x10 ? FDC_ST3_WP : 0));
	} else
		ST3 |= FDC_ST3_TS;
	drive->state = 'Q';
	fdc->result.push0(&fdc->result, ST3);	// ST3
}

static int _fdd_read_data(LPFDUDRIVES fdd, int unit, LPFDUCOMMAND cmd, int multrk_cnt) {
	LPFDU765A fdc = &fdd->owner->fdc;
	LPFDUDRIVE drive = &fdd->drives[unit];

	if (multrk_cnt == 0) {
		drive->id = cmd->id;
		drive->EOT = FALSE;
	} else if (drive->EOT)
		return -1;

	if (drive->missing_unit || !drive->media_inserted) {
		fdc->result.ST0 |= FDC_ST0_IC_ABNORM_TERM;
		fdc->result.ST1 |= FDC_ST1_MA | FDC_ST1_EN | FDC_ST1_DE | FDC_ST1_ND;
		return -1;
	}
	// find target sector to read
	LPD88SECTOR_EX sector = drive->find_sector(drive, cmd);
	if (sector == NULL) {
		fdc->result.ST0 |= FDC_ST0_IC_ABNORM_TERM;
		fdc->result.ST1 |= FDC_ST1_MA | FDC_ST1_EN | FDC_ST1_DE | FDC_ST1_ND;
		return -1;
	}
#if _DEBUG
	n80device.logger("     {READ SECTOR} [C:%d, H:%d, R:%d, N:%d]\n",
		drive->id.C, drive->id.H, drive->id.R, drive->id.N);
#endif
	drive->data_buff = drive->read_data_buff;
	drive->data_buff_index = 0;
	drive->data_size = drive->media->length((LPD88SECTOR)sector);
	drive->remain_cnt = drive->data_size;
	memset(drive->data_buff, 0xff, drive->data_size);

	if (sector->header.deleted)
		fdc->result.ST2 |= FDC_ST2_CM;
	//0x00:normal, 0x10 : normal(deleted), 0xA0 : id crc err, 0xB0 : data crc drr, 0xE0 : non addr mark, 0xF0 : non data mark
	switch (sector->header.status) {
	case 0x10:	fdc->result.ST2 |= FDC_ST2_CM;	break;
	case 0xB0:	fdc->result.ST2 |= FDC_ST2_DD;	break;
	}

	// seek id to next sector
	drive->id  = sector->header.id;
	drive->EOT = drive->forward_id(drive, (LPD88SECTOR)sector, cmd->MT, &drive->id);
	fdc->result.id = drive->id;

	drive->state = 'R';
	memcpy(drive->data_buff, sector->data, drive->data_size);
	return drive->data_size;
}

static BYTE _media_diag_buffer[DIAGSIZE];
static int _fdd_diag_data(LPFDUDRIVES fdd, int unit, LPFDUCOMMAND cmd, int multrk_cnt) {
	BYTE		buff[DIAGSIZE];
	int			ptr_arr[32];
	LPFDU765A	fdc = &fdd->owner->fdc;
	LPFDUDRIVE	drive = &fdd->drives[unit];

	// diagnostic data read done only once, so move on to result phase
	if (multrk_cnt > 0)
		return -1;

	drive->id	= cmd->id;
	drive->id.H = drive->header;// cmd->id.H == 0 && cmd->HD == 0 ? 0 : 1;
	drive->data_buff = _media_diag_buffer;
	drive->data_buff_index = 0;

	if (drive->missing_unit || !drive->media_inserted) {
		fdc->result.ST0 |= FDC_ST0_IC_ABNORM_TERM;
		fdc->result.ST1 |= FDC_ST1_MA | FDC_ST1_EN | FDC_ST1_DE | FDC_ST1_ND;
		return -1;
	}
	// get target track index
	int t = drive->cylinder * 2 + drive->header;//cmd->HD;
	if (t < 0 || t >= 164) {
		fdc->result.ST0 |= FDC_ST0_IC_ABNORM_TERM;
		fdc->result.ST1 |= FDC_ST1_MA | FDC_ST1_EN | FDC_ST1_DE | FDC_ST1_ND;
		return -1;
	}
	for (int i = 0; i < 32; ++i)	ptr_arr[i] = -1;

	// create diagnostic data (MFM only, omit MF)
	int diag_siz = 0;
	LPBYTE ptr = buff;
	// pre-amble
	for (int i = 0; i < 80; ++i)	*ptr++ = (BYTE)0x4e;			// GAP4a
	for (int i = 0; i < 12; ++i)	*ptr++ = (BYTE)0x00;			// SYNC
	for (int i = 0; i < 3; ++i)		*ptr++ = (BYTE)0xc2;			// IAM
	*ptr++ = (BYTE)0xfc;
	for (int i = 0; i < 50; ++i)	*ptr++ = (BYTE)0x4e;			// GAP1
	diag_siz = 146;

	for (int s = 0; s < 32; ++s) {
		LPD88SECTOR_EX sector = drive->media->enum_track_sector(drive->media, t, s);
		if (sector == NULL)
			break;
		if (s == (cmd->id.R - 1))
			drive->id = sector->header.id;

		// id field
		if ((diag_siz + 60) > DIAGSIZE)
			break;
		for (int i = 0; i < 12; ++i)		*ptr++ = (BYTE)0x00;	// SYNC
		for (int i = 0; i < 3; ++i)			*ptr++ = (BYTE)0xa1;	// IDAM
		*ptr++ = (BYTE)0xfe;
		*ptr++ = sector->header.id.C;
		*ptr++ = sector->header.id.H;
		*ptr++ = sector->header.id.R;
		*ptr++ = sector->header.id.N;
		for (int i = 0; i < 22 + 2; ++i)	*ptr++ = (BYTE)0x4e;	// CRC+GAP2

		// data field
		for (int i = 0; i < 12; ++i)		*ptr++ = (BYTE)0x00;	// SYNC
		for (int i = 0; i < 3; ++i)			*ptr++ = (BYTE)0xa1;	// DAM
		*ptr++ = sector->header.deleted == 0x10 ? 0xfb : 0xf8;
		diag_siz += 60;

		ptr_arr[s] = diag_siz;

		// sector data copy
		int length = drive->media->length((LPD88SECTOR)sector);
		if ((diag_siz + length) > DIAGSIZE)
			break;
		for (int i = 0; i < length; ++i)	*ptr++ = sector->data[i];
		diag_siz += length;

		// sector data end
		if ((diag_siz + 2) > DIAGSIZE)
			break;
		for (int i = 0; i < 2; ++i)			*ptr++ = (BYTE)0x00;	// CRC
		diag_siz += 2;

		// GAP3
		int gap3siz = 116;
		switch (sector->header.id.N) {
		case 0:	gap3siz = 24;	break;
		case 1: gap3siz = 54;	break;
		case 2: gap3siz = 84;	break;
		}
		if ((diag_siz + gap3siz) > DIAGSIZE)
			break;
		for (int i = 0; i < gap3siz; ++i)	*ptr++ = (BYTE)0x4e;	// GAP3
		diag_siz += gap3siz;
	}
	// fill post-amble
	if (diag_siz < DIAGSIZE) {										// GAP4b
		int gap4siz = DIAGSIZE - diag_siz;
		for (int i = 0; i < gap4siz; ++i)
			*ptr++ = (BYTE)0x4e;
	}
	// find sector data start
	int R = cmd->id.R - 1 < 32 ? cmd->id.R - 1 : 0;
	if (ptr_arr[R] == -1) {
		fdc->result.ST0 |= FDC_ST0_IC_ABNORM_TERM;
		fdc->result.ST1 |= FDC_ST1_MA | FDC_ST1_EN | FDC_ST1_DE | FDC_ST1_ND;
		return -1;
	}
	// copy to ring buffer
	ptr = _media_diag_buffer;
	for (int i = 0, j = ptr_arr[R]; i < DIAGSIZE; ++i) {
		if (j >= DIAGSIZE)	j = 0;
		*ptr++ = buff[j++];
	}
#if _DEBUG
	//n80device.logger("--------------- DIAG DATAs --------------------------\n");
	//for (int i = 0; i < DIAGSIZE; ++i) {
	//	n80device.logger("%2.2x", _media_diag_buffer[i]);
	//	if ((i % 16) == 15)
	//		n80device.logger("\n");
	//}
	//n80device.logger("\n");
#endif
	drive->data_size	= DIAGSIZE;
	drive->remain_cnt	= drive->data_size;
	fdc->result.id		= drive->id;
#if _DEBUG
	n80device.logger("     {DIAG SECTOR} [C:%d, H:%d, R:%d, N:%d]\n",
		drive->id.C, drive->id.H, drive->id.R, drive->id.N);
#endif
	drive->state = 'G';
	return drive->data_size;
}

static int _fdd_read_id(LPFDUDRIVES fdd, int unit, LPFDUCOMMAND cmd) {
	LPFDU765A fdc = &fdd->owner->fdc;
	LPFDUDRIVE drive = &fdd->drives[unit];

	drive->state = 'I';
	drive->id.C = drive->cylinder;
	drive->id.H = drive->header;// cmd->HD;

	fdc->result.ST0 = cmd->commands[1] & 0x07;
	fdc->result.ST1 = 0;
	fdc->result.ST2 = 0;
	fdc->result.let_id(&fdc->result, &drive->id);

	if (drive->missing_unit || !drive->media_inserted) {
		fdc->result.ST0 |= FDC_ST0_IC_ABNORM_TERM;
		fdc->result.ST1 |= FDC_ST1_MA | FDC_ST1_EN | FDC_ST1_DE | FDC_ST1_ND;
		return -1;
	}
	// find and read id
	int t = drive->media->find_track_index(drive->media, drive->id.C, drive->id.H);
	if (t >= 0) {
		LPD88SECTOR_EX sector = NULL, sector_1st = NULL;
		for (int s = 0; s < 32; ++s) {
			if (s <= drive->sector)
				continue;
			drive->sector = s;
			sector = drive->media->get_sector(drive->media, mktrkidx(t, s));
			if (sector == NULL)
				break;
			if (sector->header.deleted == 0x10 || sector->header.status != 0x00)
				continue;
			else
				break;
		}
		if (sector != NULL) {
			fdc->result.let_id(&fdc->result, &sector->header.id);
			return 0;
		}
		drive->sector = -1;
	}
	fdc->result.ST0 |= FDC_ST0_IC_ABNORM_TERM;
	fdc->result.ST1 |= FDC_ST1_MA | FDC_ST1_EN | FDC_ST1_DE | FDC_ST1_ND;
	return -1;
}

static void _fdd_motor_control(LPFDUDRIVES fdd, BYTE motor) {
	BYTE mon0 = motor & 0x01 ? 1 : 0;
	BYTE mon1 = motor & 0x02 ? 1 : 0;
	fdd->drives[0].motor_control(&fdd->drives[0], mon0);
	fdd->drives[1].motor_control(&fdd->drives[1], mon1);
}

static void _fdd_set_mode(LPFDUDRIVES fdd, BYTE motor) {
	BYTE td0 = motor & 0x04 ? 1 : 0;
	BYTE rv0 = motor & 0x01 ? 1 : 0;
	BYTE td1 = motor & 0x08 ? 1 : 0;
	BYTE rv1 = motor & 0x02 ? 1 : 0;
	fdd->drives[0].set_mode(&fdd->drives[0], td0, rv0);
	fdd->drives[1].set_mode(&fdd->drives[1], td1, rv1);
}

static void _fdd_reset(LPFDUDRIVES fdd) {
	for (int u = 0; u < FDC_MAX_UNIT_CNT; ++u) {
		LPFDUDRIVE drv = &fdd->drives[u];
		drv->reset(drv);
	}
}
/**--------------------------------------------------------------------------
 **		Initializer
 **------------------------------------------------------------------------*/
export void _fdu_init_drives(LPFDUDRIVES fdd, LPFDU owner) {
	fdd->owner			= owner;
	fdd->seek			= _fdd_seek;
	fdd->snsint			= _fdd_sense_interrupt_status;
	fdd->snsdrv			= _fdd_sense_drive_status;
	fdd->read			= _fdd_read_data;
	fdd->diagnostic		= _fdd_diag_data;
	fdd->read_id		= _fdd_read_id;
	fdd->reset			= _fdd_reset;
	fdd->motor_control	= _fdd_motor_control;
	fdd->set_mode		= _fdd_set_mode;
	fdd->drives_cnt = 2;
	for (int u = 0; u < FDC_MAX_UNIT_CNT; ++u) {
		LPFDUDRIVE drv = &fdd->drives[u];
		drv->owner = fdd;
		drv->init = _drv_init_drive;
		drv->missing_unit = u >= fdd->drives_cnt;
		drv->init(drv, u);
	}
}
#ifdef __cplusplus
}	// extern "C"
#endif
